from getpass import getuser
from subprocess import call
from os import path, pathsep, access, X_OK, environ, remove
from myDevices.utils.logger import exception, info, warn, error, debug
from subprocess import Popen, PIPE, check_call
from myDevices.utils.subprocess import executeCommand
from threading import RLock
from time import sleep
from socket import gethostname

APP_SETTINGS='/etc/myDevices/AppSettings.ini'

def which(program):
    """Return the path of a command executable"""

    def is_exe(fpath):
        """Return true if a file is executable"""
        return path.isfile(fpath) and access(fpath, X_OK)

    fpath, fname = path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for filepath in environ["PATH"].split(pathsep):
            smallpath = filepath.strip('"')
            exe_file = path.join(smallpath, program)
            if is_exe(exe_file):
                return exe_file
    return None


class Connection(object):
    """Class containing info about a VNC connection"""

    def __init__(self, connectionId, connectServerHost, serverPort):
        """Initialize VNC connection info"""
        self.connectionId = connectionId
        self.connectServerHost = connectServerHost
        self.serverPort = serverPort


class RemoteDesktop(object):
    """Class for creating a reverse VNC connection to a remote server"""

    def __init__(self, defaultServer, config):
        """Initialize VNC server info"""
        self.mutex = RLock()
        self.defaultServer = defaultServer
        self.isVncServerInstalled = False
        self.shareCode = None
        self.display = 2 #currently set only one display
        self.config = None
        self.serviceStarted = False
        try:
            self.config = config
            self.display = self.config.getInt('Agent', 'vncdisplay')
        except:
            pass
        self.currentHost = gethostname()
        self.currentUser = getuser()
        self.homeDir = path.expanduser('~')
        self.tightVncServerProcess = None
        info('RemoteDesktop host: {}'.format(self.currentHost))
        #HOME and USER need to be set for tightvnc to launch, though they don't seem to be set by default on Wheezy
        if 'HOME' not in environ.keys():
            environ['HOME'] = self.homeDir
        if 'USER' not in environ.keys():
            environ['USER'] = self.currentUser
        debug('Environment: {}'.format(environ))
        self.mapClients = {}
        # ThreadPool.Submit(self.CheckVncServerInstallState)

    def __del__(self):
        """Delete remote server connetion class"""
        debug('RemoteDesktop::__del__ ')
        self.Stop()

    def CheckVncServerInstallState(self):
        """Check if VNC server application is installed"""
        debug('CheckVncServerInstallState')
        if which('tightvncserver') is None:
            #vnc_install_script = "/etc/myDevices/vnc_server_install.sh"
            #try:
            #    check_call(["sudo", vnc_install_script])
            #except:
            #    error('failed installing tightvncserver')
            #    return False
            #os.remove(vnc_install_script)
            return False
        self.isVncServerInstalled = True
        return True

    def Start(self):       
        """Start the VNC server"""
        info('RemoteDesktop::Start current user: ' + self.currentUser + ' display: ' + str(self.display))
        if not which('vncserver'):
            error('Cannot start, vncserver not installed. ')
            return False
        with self.mutex:
            if len(self.mapClients) != 0:
                self.mapClients.clear()
            try:
                self.CleanupOldFiles()
                command = "vncserver :" + str(self.display)
                (output, retCode) = executeCommand(command, increaseMemoryLimit=True)
                info('VNC server command returned: ' + str(retCode) + ' output: ' + output)
                del output
                if (retCode == 1 or (retCode == 98 and not self.CheckProcessRunning())) and self.display < 7:
                    #for some reason current desktop is busy, try next one
                    self.display += 1
                    return self.Start()
                if retCode not in (0, 98):#0-success 98-already started
                    error('Failure in opening vncserver')
                    return False
            except:
                exception("RemoteDesktop::Start error")  
                return False
        if self.config:
            self.config.set('Agent', 'vncdisplay', str(self.display))
        return True

    def CleanupOldFiles(self):
        """Cleanup any leftover files from previous runs of the VNC server"""
        with self.mutex:
            try:
                if self.CheckProcessRunning():
                    return True
                else:
                    pidFilePath = self.homeDir + "/.vnc/" + self.currentHost + ":" + str(self.display) + '.pid'
                    if path.exists(pidFilePath):
                        remove(pidFilePath)
                displayFile = "/tmp/.X11-unix/X" + str(self.display)
                if path.exists(displayFile):
                    remove(displayFile)
                lockFile = "/tmp/.X" + str(self.display) + "-lock"
                if path.exists(lockFile):
                    remove(lockFile)
            except Exception as e:
                info('Exception cleaning up files {}'.format(e))
                pass

    def Stop(self):
        """Stop the VNC server"""
        debug('RemoteDesktop::Stop')
        with self.mutex:
            info('Stopping vnc server')
            command = "vncserver -kill :"+str(self.display)
            (output, retCode) = executeCommand(command)
            debug('Error Code:' + str(retCode) + ' Output: ' + output)
            if retCode == 0:
                return True
            error('Error stopping vnc server instance, return code: ' + str(retCode) + ' Output: ' + output)
            del output              
        return False

    def Connect(self, host, port):
        """Perform a reverse VNC connection to the specified host and port"""
        info('Connecting to myDevices VNC relay servers')
        command = None
        if which('tightvncconnect'): # TightVNC
            command = 'tightvncconnect -display :' + str(self.display) + ' ' + host + ':' + str(port)
        elif which('vncserver-virtual'): # RealVNC
            command = 'DISPLAY=:' + str(self.display) + ' Xvnc -connect ' + host + ':' + str(port)
        if command:
            info(command)
            (output, retCode) = executeCommand(command)
            debug('Error Code: ' + str(retCode) + ' Output: ' + output)
            if retCode != 0:
                error('Error connecting to vnc server, return code: ' + str(retCode))
                return False
            del output
            return True
        return False

    def CheckProcessRunning(self):
        """Check if the VNC server process is running"""
        pidFilePath = self.homeDir + "/.vnc/" + self.currentHost + ":" + str(self.display) + '.pid'
        info('CheckProcessRunning {}'.format(pidFilePath))
        if path.exists(pidFilePath):
            content = None
            with open(pidFilePath, 'r') as content_file:
                content = content_file.read()
            if content != None:
                try:
                    import psutil
                    if psutil.pid_exists(int(content)):
                        return True
                except:
                    exception("RemoteDesktop::CheckProcessRunning error")  
        return False

    def AddClient(self, connectionId, serverHost, serverPort):
        """Add a new remote VNC server connection"""
        info("RemoteDesktop::AddClient: " + str(len(self.mapClients)) )
        with self.mutex:
            if (self.CheckProcessRunning() == False):
                retVal = self.Start()
                if retVal == False:
                    return retVal
            connectServerHost = serverHost if serverHost != None else self.defaultServer
            self.mapClients[connectionId] = Connection(connectionId, connectServerHost, serverPort)
            return self.Connect(connectServerHost, serverPort)

    def RemoveClient(self, connectionId):
        """Remove the specified remote VNC server connection"""
        debug("RemoteDesktop::RemoveClient")
        if connectionId in self.mapClients:
            stopped = False
            del self.mapClients[connectionId]
            if len(self.mapClients) == 0:
                stopped = self.Stop()
            if stopped:
                return stopped
            return True
        return False
